#include "ferry.h"
#include "car.h"
#include "time.h"

#include <QDebug>
#include <QTimer>

Ferry::Ferry(int n, QTime time) :
	mShore(West),
	mDuration(time),
	mCapacity(n),
	mTimedOut(false)
{
	timeout();
}

void Ferry::loadCar(Car *car)
{
	QMutexLocker locker(&mMutex);
	while (mShore == East || mCarsAboard.count() >= mCapacity) {
		qDebug() << time() << "ferry full or on east shore" << car;
		mGoWait.wait(&mMutex);
		qDebug() << time() << "ferry moved (load)" << car;
	}
	mCarsAboard << car;
	mLoadWait.wakeOne();
}

Car *Ferry::unloadCar(Car *car)
{
	QMutexLocker locker(&mMutex);
	while (mShore == West) {
		qDebug() << time() << car << "awaits unloading";
		mGoWait.wait(&mMutex);
		qDebug() << time() << "ferry moved (unload)" << car;
	}
	qDebug() << time() << car << "unloaded";
	return mCarsAboard.takeAt(mCarsAboard.indexOf(car));
}

Ferry::Shore Ferry::go()
{
	QMutexLocker locker(&mMutex);
	bool ready = mShore == East;
	while (!ready) {
		qDebug() << time() << "ferry awaits for some cars";
		mLoadWait.wait(&mMutex);
		qDebug() << time() << "ferry woken up";
		if (mCarsAboard.count() == mCapacity) {
			qDebug() << time() << "ferry full";
		}
		if (mTimedOut) {
			qDebug() << time() << "ferry timed out";
		}
		ready = mCarsAboard.count() == mCapacity || (mTimedOut && !mCarsAboard.isEmpty());
		// we don't need to care if it already was false
		mTimedOut = false;
	}
	qDebug() << time() << QString("ferry leaves %1 shore")
				.arg(mShore == West ? "west" : "east");
	switch (mShore) {
		case West:
			mShore = East;
			//mUnloadWait.wakeAll();
			break;
		case East:
			mShore = West;
			//mLoadWait.wakeAll();
			break;
	}
	mGoWait.wakeAll();
	return mShore;
}

int Ferry::capacity()
{
	QMutexLocker locker(&mMutex);
	return mCapacity;
}

void Ferry::setCapacity(int n)
{
	QMutexLocker locker(&mMutex);
	mCapacity = n;
}

QTime Ferry::waitTime()
{
	QMutexLocker locker(&mMutex);
	return mDuration;
}

void Ferry::setWaitTime(QTime time)
{
	QMutexLocker locker(&mMutex);
	mDuration = time;
}

void Ferry::run()
{
	while (true) {
		Ferry::Shore shore = go();
		// make sure that car threads actually are active while the ferry is
		// docked at east shore, so that they can be unloaded
		while (shore == East && !mCarsAboard.isEmpty()) {
			QThread::yieldCurrentThread();
		}
	}
}

QTime Ferry::time() const
{
	return QTime::currentTime();
}

void Ferry::timeout()
{
	QMutexLocker locker(&mMutex);
	qDebug() << time() << "waking ferry up";
	mTimedOut = true;
	mLoadWait.wakeAll();
	QTimer::singleShot(QTime().msecsTo(mDuration), this, SLOT(timeout()));
}
